# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import subprocess
import unittest
from unittest import mock

from ansible.parsing.yaml.objects import AnsibleVaultEncryptedUnicode
import yaml

from kayobe import exception
from kayobe import utils


class TestCase(unittest.TestCase):

    maxDiff = None

    @mock.patch.object(utils, "run_command")
    def test_galaxy_role_install(self, mock_run):
        utils.galaxy_role_install("/path/to/role/file", "/path/to/roles")
        mock_run.assert_called_once_with(["ansible-galaxy", "role", "install",
                                          "--roles-path", "/path/to/roles",
                                          "--role-file", "/path/to/role/file"])

    @mock.patch.object(utils, "run_command")
    def test_galaxy_role_install_failure(self, mock_run):
        mock_run.side_effect = subprocess.CalledProcessError(1, "command")
        self.assertRaises(SystemExit,
                          utils.galaxy_role_install, "/path/to/role/file",
                          "/path/to/roles")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_yaml_file")
    def test_galaxy_collection_install(self, mock_read, mock_run):
        mock_read.return_value = {"collections": []}
        utils.galaxy_collection_install("/path/to/collection/file",
                                        "/path/to/collections")
        env = {'ANSIBLE_COLLECTIONS_SCAN_SYS_PATH': 'False'} | os.environ
        mock_run.assert_called_once_with(["ansible-galaxy", "collection",
                                          "install", "--collections-path",
                                          "/path/to/collections",
                                          "--requirements-file",
                                          "/path/to/collection/file"],
                                         env=env)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_yaml_file")
    def test_galaxy_collection_install_failure(self, mock_read, mock_run):
        mock_read.return_value = {"collections": []}
        mock_run.side_effect = subprocess.CalledProcessError(1, "command")
        self.assertRaises(SystemExit,
                          utils.galaxy_collection_install,
                          "/path/to/collection/file", "/path/to/collections")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_file")
    def test_galaxy_collection_read_failure(self, mock_read, mock_run):
        mock_read.side_effect = IOError
        self.assertRaises(SystemExit,
                          utils.galaxy_collection_install,
                          "/path/to/collection/file", "/path/to/collections")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_yaml_file")
    def test_galaxy_collection_no_collections(self, mock_read, mock_run):
        mock_read.return_value = {"roles": []}
        utils.galaxy_collection_install("/path/to/collection/file",
                                        "/path/to/collections")
        env = {'ANSIBLE_COLLECTIONS_SCAN_SYS_PATH': 'False'} | os.environ
        mock_run.assert_called_once_with(["ansible-galaxy", "collection",
                                          "install", "--collections-path",
                                          "/path/to/collections",
                                          "--requirements-file",
                                          "/path/to/collection/file"],
                                         env=env)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "read_yaml_file")
    def test_galaxy_collection_legacy_format(self, mock_read, mock_run):
        mock_read.return_value = []
        utils.galaxy_collection_install("/path/to/collection/file",
                                        "/path/to/collections")
        self.assertFalse(mock_run.called)

    @mock.patch.object(utils, "run_command")
    def test_galaxy_remove(self, mock_run):
        utils.galaxy_remove(["role1", "role2"], "/path/to/roles")
        mock_run.assert_called_once_with(["ansible-galaxy", "role", "remove",
                                          "--roles-path", "/path/to/roles",
                                          "role1", "role2"])

    @mock.patch.object(utils, "run_command")
    def test_galaxy_remove_failure(self, mock_run):
        mock_run.side_effect = subprocess.CalledProcessError(1, "command")
        self.assertRaises(SystemExit,
                          utils.galaxy_remove, ["role1", "role2"],
                          "/path/to/roles")

    @mock.patch.object(utils, "read_file")
    @mock.patch.object(yaml, "safe_load", wraps=yaml.safe_load)
    def test_read_yaml_file(self, mock_load, mock_read):
        config = """---
key1: value1
key2: value2
"""
        mock_read.return_value = config
        result = utils.read_yaml_file("/path/to/file")
        self.assertEqual(result, {"key1": "value1", "key2": "value2"})
        mock_read.assert_called_once_with("/path/to/file")
        mock_load.assert_called_once_with(config)

    @mock.patch.object(utils, "read_file")
    def test_read_yaml_file_open_failure(self, mock_read):
        mock_read.side_effect = IOError
        self.assertRaises(SystemExit, utils.read_yaml_file, "/path/to/file")

    @mock.patch.object(utils, "read_file")
    def test_read_yaml_file_not_yaml(self, mock_read):
        mock_read.return_value = "[1{!"
        self.assertRaises(SystemExit, utils.read_yaml_file, "/path/to/file")

    @mock.patch.object(utils, "read_file")
    def test_read_config_dump_yaml_file(self, mock_read):
        config = """---
key1: value1
key2: value2
"""
        mock_read.return_value = config
        result = utils.read_config_dump_yaml_file("/path/to/file")
        self.assertEqual(result, {"key1": "value1", "key2": "value2"})
        mock_read.assert_called_once_with("/path/to/file")

    @mock.patch.object(utils, "read_file")
    def test_read_config_dump_yaml_file_vaulted(self, mock_read):
        config = """---
key1: !vault |
  $ANSIBLE_VAULT;1.1;AES256
  633230623736383232323862393364323037343430393530316636363961626361393133646437
  643438663261356433656365646138666133383032376532310a63323432306431303437623637
  346236316161343635636230613838316566383933313338636237616338326439616536316639
  6334343462333062363334300a3930313762313463613537626531313230303731343365643766
  666436333037
key2: value2
key3:
  - !vault |
    $ANSIBLE_VAULT;1.1;AES256
    633230623736383232323862393364323037343430393530316636363961626361393133646437
    643438663261356433656365646138666133383032376532310a63323432306431303437623637
    346236316161343635636230613838316566383933313338636237616338326439616536316639
    6334343462333062363334300a3930313762313463613537626531313230303731343365643766
    666436333037
"""
        mock_read.return_value = config
        result = utils.read_config_dump_yaml_file("/path/to/file")
        # Can't read the value without an encryption key, so just check type.
        self.assertIsInstance(result["key1"], AnsibleVaultEncryptedUnicode)
        self.assertEqual(result["key2"], "value2")
        self.assertIsInstance(result["key3"][0], AnsibleVaultEncryptedUnicode)
        mock_read.assert_called_once_with("/path/to/file")

    @mock.patch.object(utils, "read_file")
    def test_read_config_dump_yaml_file_open_failure(self, mock_read):
        mock_read.side_effect = IOError
        self.assertRaises(SystemExit, utils.read_config_dump_yaml_file,
                          "/path/to/file")

    @mock.patch.object(utils, "read_file")
    def test_read_config_dump_yaml_file_not_yaml(self, mock_read):
        mock_read.return_value = "[1{!"
        self.assertRaises(SystemExit, utils.read_config_dump_yaml_file,
                          "/path/to/file")

    @mock.patch.object(subprocess, "check_call")
    def test_run_command(self, mock_call):
        output = utils.run_command(["command", "to", "run"])
        mock_call.assert_called_once_with(["command", "to", "run"])
        self.assertIsNone(output)

    @mock.patch.object(subprocess, "check_call")
    def test_run_command_quiet(self, mock_call):
        output = utils.run_command(["command", "to", "run"], quiet=True)
        mock_call.assert_called_once_with(["command", "to", "run"],
                                          stdout=subprocess.DEVNULL,
                                          stderr=subprocess.DEVNULL)
        self.assertIsNone(output)

    @mock.patch.object(subprocess, "check_output")
    def test_run_command_check_output(self, mock_output):
        mock_output.return_value = "command output"
        output = utils.run_command(["command", "to", "run"], check_output=True)
        mock_output.assert_called_once_with(["command", "to", "run"])
        self.assertEqual(output, "command output")

    @mock.patch.object(subprocess, "check_call")
    def test_run_command_failure(self, mock_call):
        mock_call.side_effect = subprocess.CalledProcessError(1, "command")
        self.assertRaises(subprocess.CalledProcessError, utils.run_command,
                          ["command", "to", "run"])

    def test_quote_and_escape_no_whitespace(self):
        self.assertEqual("'foo'", utils.quote_and_escape("foo"))

    def test_quote_and_escape_whitespace(self):
        self.assertEqual("'foo bar'", utils.quote_and_escape("foo bar"))

    def test_quote_and_escape_whitespace_with_quotes(self):
        self.assertEqual("'foo '\\''bar'\\'''",
                         utils.quote_and_escape("foo 'bar'"))

    def test_quote_and_escape_non_string(self):
        self.assertEqual(True, utils.quote_and_escape(True))

    def test_escape_jinja(self):
        value = "string to escape"
        expected = "{{'c3RyaW5nIHRvIGVzY2FwZQ==' | b64decode }}"
        self.assertEqual(expected, utils.escape_jinja(value))

    def test_detect_install_prefix(self):
        tmp_path = os.path.realpath('/tmp')
        path = "%s/test/local/lib/python3.6/dist-packages" % tmp_path
        expected = os.path.normpath("%s/test/local/" % tmp_path)
        result = utils._detect_install_prefix(path)
        self.assertEqual(expected, os.path.normpath(result))

    def test_detect_install_prefix_multiple_lib_components(self):
        tmp_path = os.path.realpath('/tmp')
        path = "%s/lib/test/local/lib/python3.6/dist-packages" % tmp_path
        expected = os.path.normpath("%s/lib/test/local/" % tmp_path)
        result = utils._detect_install_prefix(path)
        self.assertEqual(expected, os.path.normpath(result))

    def test_intersect_limits_no_arg_no_cli(self):
        result = utils.intersect_limits(None, None)
        self.assertEqual("", result)

    def test_intersect_limits_arg_no_cli(self):
        result = utils.intersect_limits("foo", None)
        self.assertEqual("foo", result)

    def test_intersect_limits_arg_no_cli_comma(self):
        result = utils.intersect_limits("foo,bar", None)
        self.assertEqual("foo,bar", result)

    def test_intersect_limits_arg_no_cli_colon(self):
        result = utils.intersect_limits("foo:bar", None)
        self.assertEqual("foo:bar", result)

    def test_intersect_limits_arg_no_cli_colon_and_comma(self):
        self.assertRaises(exception.Error,
                          utils.intersect_limits, "foo:bar,baz", None)

    def test_intersect_limits_no_arg_cli(self):
        result = utils.intersect_limits(None, "foo")
        self.assertEqual("foo", result)

    def test_intersect_limits_arg_and_cli(self):
        result = utils.intersect_limits("foo", "bar")
        self.assertEqual("foo:&bar", result)

    def test_intersect_limits_arg_and_cli_comma(self):
        result = utils.intersect_limits("foo,bar", "baz")
        self.assertEqual("foo,bar,&baz", result)

    def test_intersect_limits_arg_and_cli_colon(self):
        result = utils.intersect_limits("foo:bar", "baz")
        self.assertEqual("foo:bar:&baz", result)

    def test_environment_finder_with_single_environment(self):
        finder = utils.EnvironmentFinder('/etc/kayobe', 'environment-A')
        environments = finder.ordered()
        expected = ["environment-A"]
        self.assertEqual(expected, environments)

        expected = ["/etc/kayobe/environments/environment-A"]
        paths = finder.ordered_paths()
        self.assertEqual(expected, paths)

    @mock.patch.object(utils.EnvironmentFinder, "_read_metadata")
    def test_environment_finder_with_dependency_chain(self, mock_yaml):
        def yaml_replacement(path):
            if path == ("/etc/kayobe/environments/environment-C/"
                        ".kayobe-environment"):
                return {"dependencies": ["environment-A", "environment-B"]}
            if path == ("/etc/kayobe/environments/environment-B/"
                        ".kayobe-environment"):
                return {"dependencies": ["environment-A"]}
            return {}
        mock_yaml.side_effect = yaml_replacement
        finder = utils.EnvironmentFinder('/etc/kayobe', 'environment-C')
        result = finder.ordered()
        expected = ["environment-A", "environment-B", "environment-C"]
        self.assertEqual(expected, result)

        expected = ["/etc/kayobe/environments/environment-A",
                    "/etc/kayobe/environments/environment-B",
                    "/etc/kayobe/environments/environment-C"]
        paths = finder.ordered_paths()
        self.assertEqual(expected, paths)

    @mock.patch.object(utils.EnvironmentFinder, "_read_metadata")
    def test_environment_finder_with_cycle(self, mock_yaml):
        # The cycle is: C - B - C
        def yaml_replacement(path):
            if path == ("/etc/kayobe/environments/environment-C/"
                        ".kayobe-environment"):
                return {"dependencies": ["environment-A", "environment-B"]}
            if path == ("/etc/kayobe/environments/environment-B/"
                        ".kayobe-environment"):
                return {"dependencies": ["environment-A", "environment-C"]}
            return {}
        mock_yaml.side_effect = yaml_replacement
        finder = utils.EnvironmentFinder('/etc/kayobe', 'environment-C')
        self.assertRaises(exception.Error, finder.ordered)
        self.assertRaises(exception.Error, finder.ordered_paths)

    def test_environment_finder_no_environment(self):
        finder = utils.EnvironmentFinder('/etc/kayobe', None)
        self.assertEqual([], finder.ordered())
        self.assertEqual([], finder.ordered_paths())

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    def test_validate_config_path_kayobe(self, mock_readable, mock_run):
        mock_run.return_value = b"/path/to/config"
        utils.validate_config_path("/path/to/config/etc/kayobe")
        mock_run.assert_called_once_with(
            ["git", "rev-parse", "--show-toplevel"],
            check_output=True, quiet=True)
        self.assertFalse(mock_readable.called)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    def test_validate_config_path_not_a_repo(self, mock_readable, mock_run):
        mock_run.side_effect = subprocess.CalledProcessError(
            "not a repo", "command")
        utils.validate_config_path("/path/to/config/etc/kayobe")
        mock_run.assert_called_once_with(
            ["git", "rev-parse", "--show-toplevel"],
            check_output=True, quiet=True)
        self.assertFalse(mock_readable.called)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    def test_validate_config_path_no_git(self, mock_readable, mock_run):
        mock_run.side_effect = FileNotFoundError("No such file")
        utils.validate_config_path("/path/to/config/etc/kayobe")
        mock_run.assert_called_once_with(
            ["git", "rev-parse", "--show-toplevel"],
            check_output=True, quiet=True)
        self.assertFalse(mock_readable.called)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(utils, "read_file")
    def test_validate_config_path_gitreview(self, mock_read, mock_readable,
                                            mock_run):
        mock_run.return_value = b"/path/to/repo"
        mock_readable.return_value = {"result": True}
        mock_read.return_value = """
[gerrit]
project=openstack/kayobe-config.git
"""
        with self.assertLogs(level=logging.ERROR) as ctx:
            self.assertRaises(SystemExit,
                              utils.validate_config_path,
                              "/path/to/config/etc/kayobe")
            exp = ("Executing from within a different Kayobe configuration "
                   "repository is not allowed")
            assert any(exp in t for t in ctx.output)
        mock_run.assert_called_once_with(
            ["git", "rev-parse", "--show-toplevel"],
            check_output=True, quiet=True)
        mock_readable.assert_called_once_with("/path/to/repo/.gitreview")
        mock_read.assert_called_once_with("/path/to/repo/.gitreview")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    def test_validate_config_path_no_gitreview(self, mock_readable, mock_run):
        mock_run.return_value = b"/path/to/repo"
        mock_readable.return_value = {"result": False}
        utils.validate_config_path("/path/to/config/etc/kayobe")
        mock_run.assert_called_once_with(
            ["git", "rev-parse", "--show-toplevel"],
            check_output=True, quiet=True)
        mock_readable.assert_called_once_with("/path/to/repo/.gitreview")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(utils, "read_file")
    def test_validate_config_path_gitreview_no_gerrit(self, mock_read,
                                                      mock_readable, mock_run):
        mock_run.return_value = b"/path/to/repo"
        mock_readable.return_value = {"result": False}
        mock_read.return_value = """
[foo]
bar=baz
"""
        utils.validate_config_path("/path/to/config/etc/kayobe")
        mock_run.assert_called_once_with(
            ["git", "rev-parse", "--show-toplevel"],
            check_output=True, quiet=True)
        mock_readable.assert_called_once_with("/path/to/repo/.gitreview")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(utils, "read_file")
    def test_validate_config_path_gitreview_no_project(self, mock_read,
                                                       mock_readable,
                                                       mock_run):
        mock_run.return_value = b"/path/to/repo"
        mock_readable.return_value = {"result": False}
        mock_read.return_value = """
[gerrit]
bar=baz
"""
        utils.validate_config_path("/path/to/config/etc/kayobe")
        mock_run.assert_called_once_with(
            ["git", "rev-parse", "--show-toplevel"],
            check_output=True, quiet=True)
        mock_readable.assert_called_once_with("/path/to/repo/.gitreview")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(utils, "read_file")
    def test_validate_config_path_gitreview_other_project(self, mock_read,
                                                          mock_readable,
                                                          mock_run):
        mock_run.return_value = b"/path/to/repo"
        mock_readable.return_value = {"result": False}
        mock_read.return_value = """
[gerrit]
project=baz
"""
        utils.validate_config_path("/path/to/config/etc/kayobe")
        mock_run.assert_called_once_with(
            ["git", "rev-parse", "--show-toplevel"],
            check_output=True, quiet=True)
        mock_readable.assert_called_once_with("/path/to/repo/.gitreview")



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


def switch_interface_config_select_name(switch_interface_config, names):
    """Select and return all switch interfaces matching requested names.

    :param switch_interface_config: Switch interface configuration dict
    :param names: String or list of strings - interface names to match
    """
    if isinstance(names, str):
        names = [names]

    return {
        name: config
        for name, config in switch_interface_config.items()
        if name in names
    }


def switch_interface_config_select_description(switch_interface_config, descriptions):
    """Select and return all switch interfaces matching requested descriptions.

    :param switch_interface_config: Switch interface configuration dict
    :param descriptions: String or list of strings - descriptions to match
    """
    if isinstance(descriptions, str):
        descriptions = [descriptions]

    return {
        name: config
        for name, config in switch_interface_config.items()
        if (config.get('description') in descriptions and
            config.get('ngs_trunk_port', True))
    }


def switch_interface_config_select_trunk(switch_interface_config):
    """Select and return all switch interfaces which are trunk links.

    Interfaces are assumed to be trunked, unless they have a ngs_trunk_port
    item which is set to False.

    :param switch_interface_config: Switch interface configuration dict
    """
    return {
        name: config
        for name, config in switch_interface_config.items()
        if config.get('ngs_trunk_port', True)
    }


class FilterModule(object):
    """Switch filters."""

    def filters(self):
        return {
            'switch_interface_config_select_name': switch_interface_config_select_name,
            'switch_interface_config_select_description': switch_interface_config_select_description,
            'switch_interface_config_select_trunk': switch_interface_config_select_trunk,
        }



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import json
import os
import re
import sys

from cliff.command import Command
from cliff.hooks import CommandHook

from kayobe import ansible
from kayobe import environment
from kayobe import kolla_ansible
from kayobe import utils
from kayobe import vault

# This is set to an arbitrary large number to simplify the sorting logic
DEFAULT_SEQUENCE_NUMBER = sys.maxsize


def _build_playbook_list(*playbooks):
    """Return a list of names of playbook files given their basenames."""
    return [
        _get_playbook_path(playbook)
        for playbook in playbooks
    ]


def _get_playbook_path(playbook):
    """Return the absolute path of a playbook"""
    return utils.get_data_files_path("ansible", "%s.yml" % playbook)


class VaultMixin(object):
    """Mixin class for commands requiring Ansible vault."""

    def get_parser(self, prog_name):
        parser = super(VaultMixin, self).get_parser(prog_name)
        group = parser.add_argument_group("Ansible vault")
        vault.add_args(group)
        return parser


class KayobeAnsibleMixin(object):
    """Mixin class for commands running Kayobe Ansible playbooks."""

    def get_parser(self, prog_name):
        parser = super(KayobeAnsibleMixin, self).get_parser(prog_name)
        group = parser.add_argument_group("Kayobe Ansible")
        self.add_kayobe_ansible_args(group)
        return parser

    def add_kayobe_ansible_args(self, group):
        ansible.add_args(group)

    def _get_verbosity_args(self):
        """Add quietness and verbosity level arguments."""
        # Cliff's default verbosity level is 1, 0 means quiet.
        verbosity_args = {}
        if self.app.options.verbose_level:
            ansible_verbose_level = self.app.options.verbose_level - 1
            verbosity_args["verbose_level"] = ansible_verbose_level
        else:
            verbosity_args["quiet"] = True
        return verbosity_args

    def run_kayobe_playbooks(self, parsed_args, *args, **kwargs):
        kwargs.update(self._get_verbosity_args())
        return ansible.run_playbooks(parsed_args, *args, **kwargs)

    def run_kayobe_playbook(self, parsed_args, *args, **kwargs):
        kwargs.update(self._get_verbosity_args())
        return ansible.run_playbook(parsed_args, *args, **kwargs)

    def run_kayobe_config_dump(self, parsed_args, *args, **kwargs):
        kwargs.update(self._get_verbosity_args())
        return ansible.config_dump(parsed_args, *args, **kwargs)

    def generate_kolla_ansible_config(self, parsed_args, install=False,
                                      service_config=True,
                                      bifrost_config=False):
        """Generate configuration for kolla-ansible.

        :param install: If True, also install kolla-ansible.
        :param service_config: If True, generate service configuration.
        :param bifrost_config: If True, generate bifrost configuration.
        """
        # We use ignore_limit here because all of these plays execute against
        # localhost, and are typically necessary for kolla-ansible to function
        # correctly. Previously a number of people were caught out by using
        # --limit and having these plays skipped.
        tags = None if install else "config"
        playbooks = _build_playbook_list("kolla-ansible")
        self.run_kayobe_playbooks(parsed_args, playbooks, tags=tags,
                                  ignore_limit=True, check=False)
        if service_config:
            playbooks = _build_playbook_list("kolla-openstack")
            self.run_kayobe_playbooks(parsed_args, playbooks,
                                      ignore_limit=True, check=False)
        if bifrost_config:
            playbooks = _build_playbook_list("kolla-bifrost")
            self.run_kayobe_playbooks(parsed_args, playbooks,
                                      ignore_limit=True, check=False)


class KollaAnsibleMixin(object):
    """Mixin class for commands running Kolla Ansible."""

    def get_parser(self, prog_name):
        parser = super(KollaAnsibleMixin, self).get_parser(prog_name)
        group = parser.add_argument_group("Kolla Ansible")
        self.add_kolla_ansible_args(group)
        return parser

    def add_kolla_ansible_args(self, group):
        kolla_ansible.add_args(group)

    def _get_verbosity_args(self):
        """Add quietness and verbosity level arguments."""
        # Cliff's default verbosity level is 1, 0 means quiet.
        verbosity_args = {}
        if self.app.options.verbose_level:
            ansible_verbose_level = self.app.options.verbose_level - 1
            verbosity_args["verbose_level"] = ansible_verbose_level
        else:
            verbosity_args["quiet"] = True
        return verbosity_args

    def run_kolla_ansible(self, *args, **kwargs):
        kwargs.update(self._get_verbosity_args())
        return kolla_ansible.run(*args, **kwargs)

    def run_kolla_ansible_overcloud(self, *args, **kwargs):
        kwargs.update(self._get_verbosity_args())
        return kolla_ansible.run_overcloud(*args, **kwargs)

    def run_kolla_ansible_seed(self, *args, **kwargs):
        kwargs.update(self._get_verbosity_args())
        return kolla_ansible.run_seed(*args, **kwargs)


def _split_hook_sequence_number(hook):
    hook = os.path.basename(hook)
    parts = hook.split("-", 1)
    if len(parts) < 2:
        return (DEFAULT_SEQUENCE_NUMBER, hook)
    try:
        return (int(parts[0]), parts[1])
    except ValueError:
        return (DEFAULT_SEQUENCE_NUMBER, hook)


class HookDispatcher(CommandHook):
    """Runs custom playbooks before and after a command"""

# Order of calls: get_epilog, get_parser, before, after

    def __init__(self, *args, **kwargs):
        self.command = kwargs["command"]
        self.logger = self.command.app.LOG
        cmd = self.command.cmd_name
        # Replace white space with dashes for consistency with ansible
        # playbooks. Example cmd: kayobe control host bootstrap
        self.name = "-".join(cmd.split())

    def get_epilog(self):
        pass

    def get_parser(self, prog_name):
        pass

    def _find_hooks(self, env_paths, target):
        name = self.name
        # Map from hook directory path to a set of hook basenames in that path.
        hooks: {str: {str}} = {}
        for env_path in env_paths:
            path = os.path.join(env_path, "hooks", name, "%s.d" % target)
            self.logger.debug("Discovering hooks in: %s" % path)
            if not os.path.exists(path):
                continue

            hook_paths = glob.glob(os.path.join(path, "*.yml"))
            hook_basenames = {os.path.basename(hook) for hook in hook_paths}

            # Override any earlier hooks with the same basename.
            for other_hooks in hooks.values():
                other_hooks -= hook_basenames

            hooks[path] = hook_basenames

        # Return a flat list of hook paths (including directory).
        hooks = [os.path.join(path, basename)
                 for path, basenames in hooks.items()
                 for basename in basenames]
        self.logger.debug("Discovered the following hooks: %s" % hooks)
        return hooks

    def hooks(self, env_paths, target, filter):
        hooks_out = []
        if filter == "all":
            self.logger.debug("Skipping all hooks")
            return hooks_out
        hooks_in = self._find_hooks(env_paths, target)
        # Hooks can be prefixed with a sequence number to adjust running order,
        # e.g 10-my-custom-playbook.yml. Sort by sequence number.
        hooks_in = sorted(hooks_in, key=_split_hook_sequence_number)
        for hook in hooks_in:
            # Resolve symlinks so that we can reference roles.
            hook = os.path.realpath(hook)
            if filter and re.search(filter, hook):
                self.logger.debug("Skipping hook: %s", hook)
            else:
                hooks_out.append(hook)
        return hooks_out

    def run_hooks(self, parsed_args, target):
        env_paths = [parsed_args.config_path]
        environment_finder = utils.EnvironmentFinder(
            parsed_args.config_path, parsed_args.environment)
        env_paths.extend(environment_finder.ordered_paths())

        hooks = self.hooks(env_paths, target, parsed_args.skip_hooks)
        if hooks:
            self.logger.debug("Running hooks: %s" % hooks)
            self.command.run_kayobe_playbooks(parsed_args, hooks)

    def _preflight_checks(self, parsed_args):
        # NOTE(mgoddard): Currently all commands use KayobeAnsibleMixin, so
        # should have a config_path attribute, but better to be defensive.
        config_path = getattr(parsed_args, "config_path", None)
        if config_path:
            utils.validate_config_path(config_path)

    def before(self, parsed_args):
        self._preflight_checks(parsed_args)
        self.run_hooks(parsed_args, "pre")
        return parsed_args

    def after(self, parsed_args, return_code):
        if return_code == 0:
            self.run_hooks(parsed_args, "post")
        else:
            self.logger.debug("Not running hooks due to non-zero return code")
        return return_code


class ControlHostBootstrap(KayobeAnsibleMixin, KollaAnsibleMixin, VaultMixin,
                           Command):
    """Bootstrap the Kayobe control environment.

    * Downloads and installs Ansible roles from Galaxy.
    * Generates an SSH key for the Ansible control host, if one does not exist.
    * Installs kolla-ansible on the Ansible control host.
    * Generates admin-openrc.sh and public-openrc.sh files when passwords.yml
      exists.
    """

    def get_parser(self, prog_name):
        parser = super(ControlHostBootstrap, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Bootstrap")
        group.add_argument("--add-known-hosts", action='store_true',
                           help="add SSH known hosts entries for each host")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Bootstrapping Kayobe Ansible control host")
        ansible.install_galaxy_roles(parsed_args)
        ansible.install_galaxy_collections(parsed_args)
        playbooks = _build_playbook_list("bootstrap")
        self.run_kayobe_playbooks(parsed_args, playbooks, ignore_limit=True)

        passwords_exist = ansible.passwords_yml_exists(parsed_args)
        if passwords_exist:
            # Install and generate configuration - necessary for post-deploy.
            ka_tags = None
        else:
            ka_tags = "install"
        playbooks = _build_playbook_list("kolla-ansible")
        self.run_kayobe_playbooks(parsed_args, playbooks, tags=ka_tags,
                                  ignore_limit=True, check=False)

        if passwords_exist:
            # If we are bootstrapping a control host for an existing
            # environment, we should also generate the admin-openrc.sh and
            # public-openrc.sh scripts that provide admin credentials.

            self.run_kolla_ansible_overcloud(parsed_args, "post-deploy")
            # Create an environment file for accessing the public API as the
            # admin user.
            playbooks = _build_playbook_list("public-openrc")
            self.run_kayobe_playbooks(parsed_args, playbooks,
                                      ignore_limit=True)

        if parsed_args.add_known_hosts:
            self.app.LOG.debug("Adding to known_hosts")
            playbooks = _build_playbook_list("ssh-known-host")
            self.run_kayobe_playbooks(parsed_args, playbooks)


class ControlHostUpgrade(KayobeAnsibleMixin, VaultMixin, Command):
    """Upgrade the Kayobe control environment.

    * Downloads and installs updated Ansible roles from Galaxy.
    * Generates an SSH key for the Ansible control host, if one does not exist.
    * Updates kolla-ansible on the Ansible control host.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Upgrading Kayobe Ansible control host")
        # Remove roles that are no longer used. Do this before installing new
        # ones, just in case a custom role dependency includes any.
        ansible.prune_galaxy_roles(parsed_args)
        # Use force to upgrade roles and collections.
        ansible.install_galaxy_roles(parsed_args, force=True)
        ansible.install_galaxy_collections(parsed_args, force=True)
        playbooks = _build_playbook_list("bootstrap")
        self.run_kayobe_playbooks(parsed_args, playbooks, ignore_limit=True)
        playbooks = _build_playbook_list("kolla-ansible")
        self.run_kayobe_playbooks(parsed_args, playbooks, tags="install",
                                  ignore_limit=True, check=False)


class ConfigurationDump(KayobeAnsibleMixin, VaultMixin, Command):
    """Dump Kayobe configuration.

    Dumps kayobe Ansible host variables to standard output. The output may be
    filtered by selecting one or more hosts, or a specific variable.
    """

    def get_parser(self, prog_name):
        parser = super(ConfigurationDump, self).get_parser(prog_name)
        group = parser.add_argument_group("Configuration Dump")
        group.add_argument("--dump-facts", default=False,
                           help="whether to gather and dump host facts")
        group.add_argument("--host",
                           help="name of a host to dump config for")
        group.add_argument("--hosts",
                           help="name of hosts and/or groups to dump config "
                                "for")
        group.add_argument("--var-name",
                           help="name of a variable to dump")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Dumping Ansible configuration")
        hostvars = self.run_kayobe_config_dump(
            parsed_args, host=parsed_args.host, hosts=parsed_args.hosts,
            facts=parsed_args.dump_facts, var_name=parsed_args.var_name)
        try:
            json.dump(hostvars, sys.stdout, sort_keys=True, indent=4)
        except TypeError as e:
            self.app.LOG.error("Failed to JSON encode configuration: %s",
                               repr(e))
            sys.exit(1)


class PlaybookRun(KayobeAnsibleMixin, VaultMixin, Command):
    """Run a Kayobe Ansible playbook.

    Allows a single Kayobe ansible playbook to be run. For advanced users only.
    """

    def add_kayobe_ansible_args(self, group):
        super(PlaybookRun, self).add_kayobe_ansible_args(group)
        group.add_argument("playbook", nargs="+",
                           help="name of the playbook(s) to run")

    def take_action(self, parsed_args):
        self.app.LOG.debug("Running Kayobe playbook(s)")
        self.run_kayobe_playbooks(parsed_args, parsed_args.playbook)


class KollaAnsibleRun(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                      Command):
    """Run a Kolla Ansible command.

    Allows a single kolla-ansible command to be run. For advanced users only.
    """

    def add_kolla_ansible_args(self, group):
        super(KollaAnsibleRun, self).add_kolla_ansible_args(group)
        group.add_argument("--kolla-inventory-filename", default="overcloud",
                           choices=["seed", "overcloud"],
                           help="name of the kolla-ansible inventory file, "
                                "one of seed or overcloud (default "
                                "overcloud)")
        group.add_argument("command",
                           help="name of the kolla-ansible command to run")

    def take_action(self, parsed_args):
        self.app.LOG.debug("Running Kolla Ansible command")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args)

        # Run the kolla-ansible command.
        self.run_kolla_ansible(parsed_args, parsed_args.command,
                               parsed_args.kolla_inventory_filename)


class PhysicalNetworkConfigure(KayobeAnsibleMixin, VaultMixin, Command):
    """Configure a set of physical network devices."""

    def get_parser(self, prog_name):
        parser = super(PhysicalNetworkConfigure, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Physical Networking")
        group.add_argument("--group", required=True,
                           help="the Ansible group to apply configuration to")
        group.add_argument("--display", action="store_true",
                           help="display the candidate configuration and exit "
                                "without applying it")
        group.add_argument("--interface-limit",
                           help="limit the switch interfaces to be configured "
                                "by interface name")
        group.add_argument("--interface-description-limit",
                           help="limit the switch interfaces to be configured "
                                "by interface description")
        discovery = parser.add_mutually_exclusive_group()
        discovery.add_argument("--enable-discovery", action="store_true",
                               help="configure the network for hardware "
                                    "discovery")
        discovery.add_argument("--disable-discovery", action="store_true",
                               help="deconfigure the network for hardware "
                                    "discovery")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Configuring a physical network")
        extra_vars = {}
        extra_vars["physical_network_display"] = parsed_args.display
        if parsed_args.enable_discovery:
            extra_vars["physical_network_enable_discovery"] = True
        if parsed_args.disable_discovery:
            extra_vars["physical_network_disable_discovery"] = True
        if parsed_args.interface_limit:
            extra_vars["physical_network_interface_limit"] = (
                parsed_args.interface_limit)
        if parsed_args.interface_description_limit:
            extra_vars["physical_network_interface_description_limit"] = (
                parsed_args.interface_description_limit)
        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path('physical-network'),
                                 limit=parsed_args.group,
                                 extra_vars=extra_vars)


class SeedHypervisorHostConfigure(KayobeAnsibleMixin, VaultMixin, Command):
    """Configure the seed hypervisor node host OS and services.

    * Allocate IP addresses for all configured networks.
    * Add the host to SSH known hosts.
    * Configure a user account for use by kayobe for SSH access.
    * Configure proxy settings.
    * Configure package repos.
    * Configure a PyPI mirror.
    * Optionally, create a virtualenv for remote target hosts.
    * Optionally, wipe unmounted disk partitions (--wipe-disks).
    * Configure user accounts, group associations, and authorised SSH keys.
    * Configure SELinux.
    * Configure the host's network interfaces.
    * Configure a firewall.
    * Configure tuned profile.
    * Set sysctl parameters.
    * Configure IP routing and source NAT.
    * Configure timezone and ntp.
    * Optionally, configure software RAID arrays.
    * Optionally, configure encryption.
    * Configure LVM volumes.
    * Configure the host as a libvirt hypervisor.
    """

    def get_parser(self, prog_name):
        parser = super(SeedHypervisorHostConfigure, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Configuration")
        group.add_argument("--wipe-disks", action='store_true',
                           help="wipe partition and LVM data from all disks "
                                "that are not mounted. Warning: this can "
                                "result in the loss of data")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Configuring seed hypervisor host OS")
        # Allocate IP addresses.
        playbooks = _build_playbook_list("ip-allocation")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  limit="seed-hypervisor")

        kwargs = {}
        if parsed_args.wipe_disks:
            kwargs["extra_vars"] = {"wipe_disks": True}
        playbooks = _build_playbook_list("seed-hypervisor-host-configure")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  limit="seed-hypervisor", **kwargs)


class SeedHypervisorHostPackageUpdate(KayobeAnsibleMixin, VaultMixin, Command):
    """Update packages on the seed hypervisor host."""

    def get_parser(self, prog_name):
        parser = super(SeedHypervisorHostPackageUpdate, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Host Package Updates")
        group.add_argument("--packages", required=True,
                           help="List of packages to update. Use '*' to "
                                "update all packages.")
        group.add_argument("--security", action='store_true',
                           help="Only install updates that have been marked "
                                "security related.")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Updating seed hypervisor host packages")
        extra_vars = {
            "host_package_update_packages": parsed_args.packages,
            "host_package_update_security": parsed_args.security,
        }
        playbooks = _build_playbook_list("host-package-update")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  limit="seed-hypervisor",
                                  extra_vars=extra_vars)


class SeedHypervisorHostCommandRun(KayobeAnsibleMixin, VaultMixin, Command):
    """Run command on the seed hypervisor host."""

    def get_parser(self, prog_name):
        parser = super(SeedHypervisorHostCommandRun, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Host Command Run")
        group.add_argument("--command", required=True,
                           help="Command to run (required).")
        group.add_argument("--show-output", action='store_true',
                           help="Show command output")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Run command on seed hypervisor host")
        extra_vars = {
            "host_command_to_run": utils.escape_jinja(parsed_args.command),
            "show_output": parsed_args.show_output}
        playbooks = _build_playbook_list("host-command-run")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  limit="seed-hypervisor",
                                  extra_vars=extra_vars)


class SeedHypervisorHostUpgrade(KayobeAnsibleMixin, VaultMixin, Command):
    """Upgrade the seed hypervisor host services.

    Performs the changes necessary to make the host services suitable for the
    configured OpenStack release.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Upgrading seed hypervisor host services")
        playbooks = _build_playbook_list("kayobe-target-venv")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  limit="seed-hypervisor")


class SeedVMProvision(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                      Command):
    """Provision the seed VM.

    * Allocate IP addresses for all configured networks.
    * Provision a virtual machine using libvirt.
    * Configure the kolla-ansible inventory for the seed VM.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Provisioning seed VM")
        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path("ip-allocation"),
                                 limit="seed")
        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path("seed-vm-provision"))
        # Now populate the Kolla Ansible inventory.
        self.generate_kolla_ansible_config(parsed_args, service_config=False)


class SeedVMDeprovision(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                        Command):
    """Deprovision the seed VM.

    This will destroy the seed VM and all associated volumes.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Deprovisioning seed VM")
        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path("seed-vm-deprovision"))


class SeedHostConfigure(KayobeAnsibleMixin, VaultMixin, Command):
    """Configure the seed node host OS and services.

    * Allocate IP addresses for all configured networks.
    * Add the host to SSH known hosts.
    * Configure a user account for use by kayobe for SSH access.
    * Configure proxy settings.
    * Configure package repos.
    * Configure a PyPI mirror.
    * Optionally, create a virtualenv for remote target hosts.
    * Optionally, wipe unmounted disk partitions (--wipe-disks).
    * Configure user accounts, group associations, and authorised SSH keys.
    * Configure SELinux.
    * Configure the host's network interfaces.
    * Configure a firewall.
    * Configure tuned profile.
    * Set sysctl parameters.
    * Configure IP routing and source NAT.
    * Disable bootstrap interface configuration.
    * Configure timezone and ntp.
    * Optionally, configure software RAID arrays.
    * Optionally, configure encryption.
    * Configure LVM volumes.
    * Optionally, create a virtualenv for kolla-ansible.
    * Configure a user account for kolla-ansible.
    * Configure Docker engine.
    * Optionally, deploy a Docker Registry.
    """

    def get_parser(self, prog_name):
        parser = super(SeedHostConfigure, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Configuration")
        group.add_argument("--wipe-disks", action='store_true',
                           help="wipe partition and LVM data from all disks "
                                "that are not mounted. Warning: this can "
                                "result in the loss of data")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Configuring seed host OS")

        # Allocate IP addresses.
        playbooks = _build_playbook_list("ip-allocation")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="seed")

        # Run kayobe playbooks.
        extra_vars = {"kayobe_action": "deploy"}
        if parsed_args.wipe_disks:
            extra_vars["wipe_disks"] = True
        playbooks = _build_playbook_list("seed-host-configure")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="seed",
                                  extra_vars=extra_vars)


class SeedHostPackageUpdate(KayobeAnsibleMixin, VaultMixin, Command):
    """Update packages on the seed host."""

    def get_parser(self, prog_name):
        parser = super(SeedHostPackageUpdate, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Package Updates")
        group.add_argument("--packages", required=True,
                           help="List of packages to update. Use '*' to "
                                "update all packages.")
        group.add_argument("--security", action='store_true',
                           help="Only install updates that have been marked "
                                "security related.")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Updating seed host packages")
        extra_vars = {
            "host_package_update_packages": parsed_args.packages,
            "host_package_update_security": parsed_args.security,
        }
        playbooks = _build_playbook_list("host-package-update")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="seed",
                                  extra_vars=extra_vars)


class SeedHostCommandRun(KayobeAnsibleMixin, VaultMixin, Command):
    """Run command on the seed host."""

    def get_parser(self, prog_name):
        parser = super(SeedHostCommandRun, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Command Run")
        group.add_argument("--command", required=True,
                           help="Command to run (required).")
        group.add_argument("--show-output", action='store_true',
                           help="Show command output")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Run command on seed host")
        extra_vars = {
            "host_command_to_run": utils.escape_jinja(parsed_args.command),
            "show_output": parsed_args.show_output}
        playbooks = _build_playbook_list("host-command-run")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="seed",
                                  extra_vars=extra_vars)


class SeedHostUpgrade(KayobeAnsibleMixin, VaultMixin, Command):
    """Upgrade the seed host services.

    Performs the changes necessary to make the host services suitable for the
    configured OpenStack release.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Upgrading seed host services")
        playbooks = _build_playbook_list("seed-host-upgrade")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="seed")


class SeedServiceDeploy(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                        Command):
    """Deploy the seed services.

    * Deploys user defined containers
    * Configures kolla-ansible.
    * Configures the bifrost service.
    * Deploys the bifrost container using kolla-ansible.
    * Builds disk images for the overcloud hosts using Diskimage Builder (DIB).
    * Configures ironic inspector introspection rules in the bifrost inspector
      service.
    * When enabled, configures a Bare Metal Provisioning (BMP) environment for
      Dell Force10 switches, hosted by the bifrost dnsmasq and nginx services.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Deploying seed services")
        playbooks = _build_playbook_list(
            "seed-manage-containers")
        extra_vars = {"kayobe_action": "deploy"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)
        self.generate_kolla_ansible_config(parsed_args, service_config=False,
                                           bifrost_config=True)

        self.run_kolla_ansible_seed(parsed_args, "deploy-bifrost")
        playbooks = _build_playbook_list(
            "seed-credentials",
            "seed-introspection-rules",
            "dell-switch-bmp")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class SeedServiceDestroy(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                         Command):
    """Destroy the seed services.

    * Destroys user defined containers
    * Destroys kolla deployed containers
    * Destroys docker registry
    """

    def take_action(self, parsed_args):
        if not parsed_args.yes_i_really_really_mean_it:
            self.app.LOG.error("This will permanently destroy all services "
                               "and data. Specify "
                               "--yes-i-really-really-mean-it to confirm that "
                               "you understand this.")
            sys.exit(1)
        self.app.LOG.debug("Destroying seed services")
        self.generate_kolla_ansible_config(parsed_args, service_config=False,
                                           bifrost_config=False)
        extra_args = ["--yes-i-really-really-mean-it"]
        self.run_kolla_ansible_seed(parsed_args, "destroy",
                                    extra_args=extra_args)

        extra_vars = {"kayobe_action": "destroy"}
        playbooks = _build_playbook_list(
            "seed-manage-containers",
            "docker-registry")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)

    def get_parser(self, prog_name):
        parser = super(SeedServiceDestroy, self).get_parser(prog_name)
        group = parser.add_argument_group("Services")
        group.add_argument("--yes-i-really-really-mean-it",
                           action='store_true',
                           help="confirm that you understand that this will "
                                "permanently destroy all services and data.")
        return parser


class SeedServiceUpgrade(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                         Command):
    """Upgrade the seed services.

    * Deploys user defined containers
    * Configures kolla-ansible.
    * Configures the bifrost service.
    * Prepares the bifrost service for an upgrade.
    * Deploys the bifrost container using kolla-ansible.
    * Builds disk images for the overcloud hosts using Diskimage Builder (DIB).
    * Configures ironic inspector introspection rules in the bifrost inspector
      service.
    * When enabled, configures a Bare Metal Provisioning (BMP) environment for
      Dell Force10 switches, hosted by the bifrost dnsmasq and nginx services.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Upgrading seed services")
        playbooks = _build_playbook_list(
            "seed-manage-containers")
        extra_vars = {"kayobe_action": "deploy"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)
        self.generate_kolla_ansible_config(parsed_args, service_config=False,
                                           bifrost_config=True)

        playbooks = _build_playbook_list(
            "seed-service-upgrade-prep")
        self.run_kayobe_playbooks(parsed_args, playbooks)
        self.run_kolla_ansible_seed(parsed_args, "upgrade-bifrost")
        playbooks = _build_playbook_list(
            "seed-credentials",
            "seed-introspection-rules",
            "dell-switch-bmp")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class SeedContainerImageBuild(KayobeAnsibleMixin, VaultMixin, Command):
    """Build the seed container images.

    * Installs and configures kolla build environment on the seed.
    * Builds container images for the seed services.
    """

    def get_parser(self, prog_name):
        parser = super(SeedContainerImageBuild, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Container Image Build")
        group.add_argument("--nocache", action="store_true",
                           help="whether to not use cache")
        group.add_argument("--push", action="store_true",
                           help="whether to push images to a registry after "
                                "building")
        group.add_argument("regex", nargs='*',
                           help="regular expression matching names of images "
                                "to build. Builds all images if unspecified")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Building seed container images")
        playbooks = _build_playbook_list(
            "container-image-builders-check", "kolla-build",
            "container-image-build")
        extra_vars = {
            "nocache": parsed_args.nocache,
            "push_images": parsed_args.push
        }
        if parsed_args.regex:
            regexes = " ".join(parsed_args.regex)
            extra_vars["container_image_regexes"] = regexes
        else:
            extra_vars["container_image_sets"] = (
                "{{ seed_container_image_sets }}")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class SeedDeploymentImageBuild(KayobeAnsibleMixin, VaultMixin, Command):
    """Build the seed deployment kernel and ramdisk images.

    Builds Ironic Python Agent (IPA) deployment images using Diskimage Builder
    (DIB) for use when provisioning and inspecting the overcloud hosts.
    """

    def get_parser(self, prog_name):
        parser = super(SeedDeploymentImageBuild, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Deployment Image Build")
        group.add_argument("--force-rebuild", action="store_true",
                           help="whether to force rebuilding the images")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Building seed deployment images")
        playbooks = _build_playbook_list("seed-ipa-build")
        extra_vars = {}
        if parsed_args.force_rebuild:
            extra_vars["ipa_image_force_rebuild"] = True
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class InfraVMProvision(KayobeAnsibleMixin, VaultMixin, Command):
    """Provisions infra virtual machines

    * Allocate IP addresses for all configured networks.
    * Provision a virtual machine using libvirt.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Provisioning infra VMs")
        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path("ip-allocation"),
                                 limit="infra-vms")

        limit_arg = utils.intersect_limits(parsed_args.limit, "infra-vms")
        # We want the limit to affect one play only. To do this we use a
        # variable to override the hosts list instead of using limit.
        extra_vars = {
            "infra_vm_limit": limit_arg
        }

        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path("infra-vm-provision"),
                                 ignore_limit=True, extra_vars=extra_vars)


class InfraVMDeprovision(KayobeAnsibleMixin, VaultMixin, Command):
    """Deprovisions infra virtual machines.

    This will destroy all infra VMs and all associated volumes.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Deprovisioning infra VMs")
        # We want the limit to affect one play only. To do this we use a
        # variable to override the hosts list instead of using limit.
        limit_arg = utils.intersect_limits(parsed_args.limit, "infra-vms")
        extra_vars = {
            "infra_vm_limit": limit_arg
        }

        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path("infra-vm-deprovision"),
                                 ignore_limit=True, extra_vars=extra_vars)


class InfraVMHostConfigure(KayobeAnsibleMixin, VaultMixin, Command):
    """Configure the infra VMs host OS and services.

    * Allocate IP addresses for all configured networks.
    * Add the host to SSH known hosts.
    * Configure a user account for use by kayobe for SSH access.
    * Configure proxy settings.
    * Configure package repos.
    * Configure a PyPI mirror.
    * Optionally, create a virtualenv for remote target hosts.
    * Optionally, wipe unmounted disk partitions (--wipe-disks).
    * Configure user accounts, group associations, and authorised SSH keys.
    * Configure SELinux.
    * Configure the host's network interfaces.
    * Configure a firewall.
    * Configure tuned profile.
    * Set sysctl parameters.
    * Disable bootstrap interface configuration.
    * Configure timezone and ntp.
    * Optionally, configure software RAID arrays.
    * Optionally, configure encryption.
    * Configure LVM volumes.
    """

    def get_parser(self, prog_name):
        parser = super(InfraVMHostConfigure, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Configuration")
        group.add_argument("--wipe-disks", action='store_true',
                           help="wipe partition and LVM data from all disks "
                                "that are not mounted. Warning: this can "
                                "result in the loss of data")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Configuring Infra VMs host OS")

        # Allocate IP addresses.
        playbooks = _build_playbook_list("ip-allocation")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="infra-vms")

        # Kayobe playbooks.
        kwargs = {}
        if parsed_args.wipe_disks:
            kwargs["extra_vars"] = {"wipe_disks": True}
        playbooks = _build_playbook_list("infra-vm-host-configure")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="infra-vms",
                                  **kwargs)


class InfraVMHostPackageUpdate(KayobeAnsibleMixin, VaultMixin, Command):
    """Update packages on the infra VMs."""

    def get_parser(self, prog_name):
        parser = super(InfraVMHostPackageUpdate, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Package Updates")
        group.add_argument("--packages", required=True,
                           help="List of packages to update. Use '*' to "
                                "update all packages.")
        group.add_argument("--security", action='store_true',
                           help="Only install updates that have been marked "
                                "security related.")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Updating infra vm host packages")
        extra_vars = {
            "host_package_update_packages": parsed_args.packages,
            "host_package_update_security": parsed_args.security,
        }
        playbooks = _build_playbook_list("host-package-update")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="infra-vms",
                                  extra_vars=extra_vars)


class InfraVMHostCommandRun(KayobeAnsibleMixin, VaultMixin, Command):
    """Run command on the infra VMs."""

    def get_parser(self, prog_name):
        parser = super(InfraVMHostCommandRun, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Command Run")
        group.add_argument("--command", required=True,
                           help="Command to run (required).")
        group.add_argument("--show-output", action='store_true',
                           help="Show command output")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Run command on infra VM hosts")
        extra_vars = {
            "host_command_to_run": utils.escape_jinja(parsed_args.command),
            "show_output": parsed_args.show_output}
        playbooks = _build_playbook_list("host-command-run")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="infra-vms",
                                  extra_vars=extra_vars)


class InfraVMHostUpgrade(KayobeAnsibleMixin, VaultMixin, Command):
    """Upgrade the infra VM host services.

    Performs the changes necessary to make the host services suitable for the
    configured OpenStack release.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Upgrading the infra-vm host services")
        playbooks = _build_playbook_list("kayobe-target-venv")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  limit="infra-vms")


class InfraVMServiceDeploy(KayobeAnsibleMixin, VaultMixin,
                           Command):
    """Run hooks for infra structure services."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Running no-op Infra VM service deploy")


class OvercloudInventoryDiscover(KayobeAnsibleMixin, VaultMixin, Command):
    """Discover the overcloud inventory from the seed's Ironic service.

    * Query the ironic inventory on the seed, and use this to populate kayobe's
      ansible inventory.
    * Allocate IP addresses for all configured networks.
    * Update the kolla-ansible configuration for the new overcloud hosts.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Discovering overcloud inventory")
        # Run the inventory discovery playbook separately, else the discovered
        # hosts will not be present in the following playbooks in which they
        # are used to populate other inventories.
        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path(
                                     "overcloud-inventory-discover"))
        # If necessary, allocate IP addresses for the discovered hosts.
        self.run_kayobe_playbook(parsed_args,
                                 _get_playbook_path("ip-allocation"),
                                 limit="overcloud")
        # Now populate the Kolla Ansible inventory.
        self.generate_kolla_ansible_config(parsed_args, service_config=False)


class OvercloudIntrospectionDataSave(KayobeAnsibleMixin, VaultMixin, Command):
    """Save hardware introspection data for the overcloud.

    Save hardware introspection data from the seed's ironic inspector service
    to the Ansible control host.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudIntrospectionDataSave, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Introspection data")
        # Defaults for these are applied in the playbook.
        group.add_argument("--output-dir", type=str,
                           help="Path to directory in which to save "
                                "introspection data. Default: "
                                "$PWD/overcloud-introspection-data")
        group.add_argument("--output-format", type=str,
                           help="Format in which to save output data. One of "
                                "JSON or YAML. Default: JSON",
                           choices=["JSON", "YAML"])
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Saving introspection data")
        extra_vars = {}
        if parsed_args.output_dir:
            extra_vars['output_dir'] = parsed_args.output_dir
        if parsed_args.output_format:
            extra_vars['output_format'] = parsed_args.output_format
        playbooks = _build_playbook_list("kolla-bifrost-hostvars",
                                         "overcloud-introspection-data-save")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class OvercloudBIOSRAIDConfigure(KayobeAnsibleMixin, VaultMixin, Command):
    """Configure BIOS and RAID for the overcloud hosts."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Configure overcloud BIOS and RAID")
        playbooks = _build_playbook_list("overcloud-bios-raid")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class OvercloudHardwareInspect(KayobeAnsibleMixin, VaultMixin, Command):
    """Inspect the overcloud hardware using ironic inspector.

    Perform hardware inspection of existing ironic nodes in the seed's
    ironic inventory.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Inspecting overcloud")
        playbooks = _build_playbook_list("kolla-bifrost-hostvars",
                                         "overcloud-hardware-inspect")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class OvercloudProvision(KayobeAnsibleMixin, VaultMixin, Command):
    """Provision the overcloud.

    Provision the overcloud hosts using the seed host's bifrost service. This
    will image the hosts and perform some minimal network configuration using
    glean/simple-init.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Provisioning overcloud")
        playbooks = _build_playbook_list("kolla-bifrost-hostvars",
                                         "overcloud-provision")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class OvercloudDeprovision(KayobeAnsibleMixin, VaultMixin, Command):
    """Deprovision the overcloud.

    Deprovision the overcloud hosts using the seed host's bifrost service. This
    will clear the instance state of the nodes from the seed's ironic service
    and power them off.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Deprovisioning overcloud")
        playbooks = _build_playbook_list("kolla-bifrost-hostvars",
                                         "overcloud-deprovision")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class OvercloudFactsGather(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                           Command):
    """Gather facts for Kayobe and Kolla Ansible."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Gathering overcloud host facts")

        # Gather facts for Kayobe.
        playbooks = _build_playbook_list("overcloud-facts-gather")
        self.run_kayobe_playbooks(parsed_args, playbooks)

        # Gather facts for Kolla Ansible.
        self.generate_kolla_ansible_config(parsed_args, service_config=False)
        self.run_kolla_ansible_overcloud(parsed_args, "gather-facts")


class OvercloudHostConfigure(KayobeAnsibleMixin, VaultMixin, Command):
    """Configure the overcloud host OS and services.

    * Allocate IP addresses for all configured networks.
    * Add the host to SSH known hosts.
    * Configure a user account for use by kayobe for SSH access.
    * Configure proxy settings.
    * Configure package repos.
    * Configure a PyPI mirror.
    * Optionally, create a virtualenv for remote target hosts.
    * Optionally, wipe unmounted disk partitions (--wipe-disks).
    * Configure user accounts, group associations, and authorised SSH keys.
    * Configure SELinux.
    * Configure the host's network interfaces.
    * Configure a firewall.
    * Configure tuned profile.
    * Set sysctl parameters.
    * Disable bootstrap interface configuration.
    * Configure timezone and ntp.
    * Optionally, configure software RAID arrays.
    * Optionally, configure encryption.
    * Configure LVM volumes.
    * Optionally, create a virtualenv for kolla-ansible.
    * Configure a user account for kolla-ansible.
    * Configure Docker engine.
    * Configure libvirt.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudHostConfigure, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Configuration")
        group.add_argument("--wipe-disks", action='store_true',
                           help="wipe partition and LVM data from all disks "
                                "that are not mounted. Warning: this can "
                                "result in the loss of data")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Configuring overcloud host OS")

        # Allocate IP addresses.
        playbooks = _build_playbook_list("ip-allocation")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="overcloud")

        # Kayobe playbooks.
        kwargs = {}
        if parsed_args.wipe_disks:
            kwargs["extra_vars"] = {"wipe_disks": True}
        playbooks = _build_playbook_list("overcloud-host-configure")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="overcloud",
                                  **kwargs)


class OvercloudHostPackageUpdate(KayobeAnsibleMixin, VaultMixin, Command):
    """Update packages on the overcloud hosts."""

    def get_parser(self, prog_name):
        parser = super(OvercloudHostPackageUpdate, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Package Updates")
        group.add_argument("--packages", required=True,
                           help="List of packages to update. Use '*' to "
                                "update all packages.")
        group.add_argument("--security", action='store_true',
                           help="Only install updates that have been marked "
                                "security related.")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Updating overcloud host packages")
        extra_vars = {
            "host_package_update_packages": parsed_args.packages,
            "host_package_update_security": parsed_args.security,
        }
        playbooks = _build_playbook_list("host-package-update")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="overcloud",
                                  extra_vars=extra_vars)


class OvercloudHostCommandRun(KayobeAnsibleMixin, VaultMixin, Command):
    """Run command on the overcloud host."""

    def get_parser(self, prog_name):
        parser = super(OvercloudHostCommandRun, self).get_parser(prog_name)
        group = parser.add_argument_group("Host Command Run")
        group.add_argument("--command", required=True,
                           help="Command to run (required).")
        group.add_argument("--show-output", action='store_true',
                           help="Show command output")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Run command on overcloud host")
        extra_vars = {
            "host_command_to_run": utils.escape_jinja(parsed_args.command),
            "show_output": parsed_args.show_output}
        playbooks = _build_playbook_list("host-command-run")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="overcloud",
                                  extra_vars=extra_vars)


class OvercloudHostUpgrade(KayobeAnsibleMixin, VaultMixin, Command):
    """Upgrade the overcloud host services.

    Performs the changes necessary to make the host services suitable for the
    configured OpenStack release.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Upgrading overcloud host services")
        playbooks = _build_playbook_list("overcloud-host-upgrade")
        self.run_kayobe_playbooks(parsed_args, playbooks, limit="overcloud")


class OvercloudDatabaseBackup(KollaAnsibleMixin, KayobeAnsibleMixin,
                              VaultMixin, Command):
    """Backup the overcloud database."""

    def get_parser(self, prog_name):
        parser = super(OvercloudDatabaseBackup, self).get_parser(prog_name)
        group = parser.add_argument_group("Overcloud Database Backup")
        group.add_argument("--incremental", action='store_true',
                           help="Whether to perform an incremental database "
                                "backup. Default is false.")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Performing overcloud database backup")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args, service_config=False)

        extra_args = []
        if parsed_args.incremental:
            extra_args.append('--incremental')
        self.run_kolla_ansible_overcloud(parsed_args, "mariadb-backup",
                                         extra_args=extra_args)


class OvercloudDatabaseRecover(KollaAnsibleMixin, KayobeAnsibleMixin,
                               VaultMixin, Command):
    """Recover the overcloud database."""

    def get_parser(self, prog_name):
        parser = super(OvercloudDatabaseRecover, self).get_parser(prog_name)
        group = parser.add_argument_group("Overcloud Database Recovery")
        group.add_argument("--force-recovery-host",
                           help="Name of a host to use to perform the "
                                "recovery. By default kolla-ansible will "
                                "automatically determine which host to use, "
                                "and this option should not be used.")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Performing overcloud database recovery")
        extra_vars = {}

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args, service_config=True)

        if parsed_args.force_recovery_host:
            extra_vars['mariadb_recover_inventory_name'] = (
                parsed_args.force_recovery_host)
        self.run_kolla_ansible_overcloud(parsed_args, "mariadb-recovery",
                                         extra_vars=extra_vars)


class OvercloudServiceConfigurationGenerate(KayobeAnsibleMixin,
                                            KollaAnsibleMixin, VaultMixin,
                                            Command):
    """Generate the overcloud service configuration files.

    Generates kolla-ansible configuration for the OpenStack control plane
    services, without pushing that configuration to the running containers.
    This can be used to generate a candidate configuration set for comparison
    with the existing configuration. It is recommended to use a directory other
    than /etc/kolla for --node-config-dir, to ensure that the running
    containers are not affected.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceConfigurationGenerate,
                       self).get_parser(prog_name)
        group = parser.add_argument_group("Service Configuration Generate")
        group.add_argument("--node-config-dir", required=True,
                           help="the directory to store the config files on "
                                "the remote node (required)")
        group.add_argument("--skip-prechecks", action='store_true',
                           help="skip the kolla-ansible prechecks command")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Generating overcloud service configuration")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args)

        # Run kolla-ansible prechecks before deployment.
        if not parsed_args.skip_prechecks:
            self.run_kolla_ansible_overcloud(parsed_args, "prechecks")

        # Generate the configuration.
        extra_vars = {}
        if parsed_args.node_config_dir:
            extra_vars["node_config_directory"] = parsed_args.node_config_dir
        self.run_kolla_ansible_overcloud(parsed_args, "genconfig",
                                         extra_vars=extra_vars)


class OvercloudServiceConfigurationValidate(KayobeAnsibleMixin,
                                            KollaAnsibleMixin, VaultMixin,
                                            Command):
    """Runs oslo config validator on the OpenStack control plane services."""

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceConfigurationValidate,
                       self).get_parser(prog_name)
        group = parser.add_argument_group("Service Configuration Validate")
        group.add_argument("--output-dir", required=True,
                           help="the directory to store the results of running"
                                " the config validator (required)")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Validating overcloud service configuration")
        extra_vars = {}
        if parsed_args.output_dir:
            extra_vars[
                "service_config_validate_output_dir"] = parsed_args.output_dir
        self.run_kolla_ansible_overcloud(parsed_args, "validate-config",
                                         extra_vars=extra_vars)


class OvercloudServiceConfigurationSave(KayobeAnsibleMixin, VaultMixin,
                                        Command):
    """Gather and save the overcloud service configuration files.

    This can be used to collect the running configuration for inspection (the
    default) or a candidate configuration generated via 'kayobe overcloud
    service configuration generate', for comparision with another configuration
    set.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceConfigurationSave, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Service Configuration Save")
        group.add_argument("--exclude",
                           help="optional comma-separated list of patterns "
                                "matching filenames to exclude")
        group.add_argument("--include",
                           help="optional comma-separated list of patterns "
                                "matching filenames to include")
        group.add_argument("--node-config-dir",
                           help="the directory to store the config files on "
                                "the remote node (default /etc/kolla)")
        group.add_argument("--output-dir",
                           help="path to a directory in which to save "
                                "configuration")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Saving overcloud service configuration")
        playbooks = _build_playbook_list("overcloud-service-config-save")
        extra_vars = {}
        if parsed_args.exclude:
            extra_vars["exclude_patterns"] = parsed_args.exclude
        if parsed_args.include:
            extra_vars["include_patterns"] = parsed_args.include
        if parsed_args.output_dir:
            extra_vars["config_save_path"] = parsed_args.output_dir
        if parsed_args.node_config_dir:
            extra_vars["node_config_directory"] = parsed_args.node_config_dir
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class OvercloudServiceDeploy(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                             Command):
    """Deploy the overcloud services.

    * Configure kolla-ansible.
    * Configure overcloud services in kolla-ansible.
    * Perform kolla-ansible prechecks to verify the system state for
      deployment.
    * Perform a kolla-ansible deployment of the overcloud services.
    * Configure and deploy kayobe extra services.
    * Generate openrc files for the admin user.

    This can be used in conjunction with the --tags and --kolla-tags arguments
    to deploy specific services.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceDeploy, self).get_parser(prog_name)
        group = parser.add_argument_group("Service Deployment")
        group.add_argument("--skip-prechecks", action='store_true',
                           help="skip the kolla-ansible prechecks command")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Deploying overcloud services")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args)

        # Run kolla-ansible prechecks before deployment.
        if not parsed_args.skip_prechecks:
            self.run_kolla_ansible_overcloud(parsed_args, "prechecks")

        # Perform the kolla-ansible deployment.
        self.run_kolla_ansible_overcloud(parsed_args, "deploy")

        # Deploy kayobe extra services.
        playbooks = _build_playbook_list("overcloud-extras")
        extra_vars = {"kayobe_action": "deploy"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars, limit="overcloud")

        # Post-deployment configuration.
        self.run_kolla_ansible_overcloud(parsed_args, "post-deploy")
        # Create an environment file for accessing the public API as the admin
        # user.
        playbooks = _build_playbook_list("public-openrc")
        self.run_kayobe_playbooks(parsed_args, playbooks, ignore_limit=True)


class OvercloudServiceDeployContainers(KollaAnsibleMixin, KayobeAnsibleMixin,
                                       VaultMixin, Command):
    """Deploy the overcloud services without updating configuration.

    * Configure kolla-ansible.
    * Configure overcloud services in kolla-ansible.
    * Perform kolla-ansible prechecks to verify the system state for
      deployment.
    * Perform a kolla-ansible deployment of the overcloud service containers.
    * Configure and deploy kayobe extra services.

    This can be used in conjunction with the --tags and --kolla-tags arguments
    to deploy specific services.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceDeployContainers, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Service Deployment")
        group.add_argument("--skip-prechecks", action='store_true',
                           help="skip the kolla-ansible prechecks command")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Deploying overcloud services (containers only)")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args)

        # Run kolla-ansible prechecks before deployment.
        if not parsed_args.skip_prechecks:
            self.run_kolla_ansible_overcloud(parsed_args, "prechecks")

        # Perform the kolla-ansible deployment.
        self.run_kolla_ansible_overcloud(parsed_args, "manage-containers")

        # Deploy kayobe extra services.
        playbooks = _build_playbook_list("overcloud-extras")
        extra_vars = {"kayobe_action": "deploy"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars, limit="overcloud")


class OvercloudServicePrechecks(KollaAnsibleMixin, KayobeAnsibleMixin,
                                VaultMixin, Command):
    """Run prechecks against overcloud services.

    * Configure kolla-ansible.
    * Configure overcloud services in kolla-ansible.
    * Perform kolla-ansible prechecks to verify the system state for
      deployment.

    This can be used in conjunction with the --tags and --kolla-tags arguments
    to check specific services.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Running overcloud prechecks")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args)

        # Run the kolla-ansible prechecks.
        self.run_kolla_ansible_overcloud(parsed_args, "prechecks")


class OvercloudServiceReconfigure(KollaAnsibleMixin, KayobeAnsibleMixin,
                                  VaultMixin, Command):
    """Reconfigure the overcloud services.

    * Configure kolla-ansible.
    * Configure overcloud services in kolla-ansible.
    * Perform kolla-ansible prechecks to verify the system state for
      deployment.
    * Perform a kolla-ansible reconfiguration of the overcloud services.
    * Configure and deploy kayobe extra services.
    * Generate openrc files for the admin user.

    This can be used in conjunction with the --tags and --kolla-tags arguments
    to reconfigure specific services.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceReconfigure, self).get_parser(prog_name)
        group = parser.add_argument_group("Service Reconfiguration")
        group.add_argument("--skip-prechecks", action='store_true',
                           help="skip the kolla-ansible prechecks command")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Reconfiguring overcloud services")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args)

        # Run kolla-ansible prechecks before reconfiguration.
        if not parsed_args.skip_prechecks:
            self.run_kolla_ansible_overcloud(parsed_args, "prechecks")

        # Perform the kolla-ansible reconfiguration.
        self.run_kolla_ansible_overcloud(parsed_args, "reconfigure")

        # Reconfigure kayobe extra services.
        playbooks = _build_playbook_list("overcloud-extras")
        extra_vars = {"kayobe_action": "reconfigure"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars, limit="overcloud")

        # Post-deployment configuration.
        self.run_kolla_ansible_overcloud(parsed_args, "post-deploy")
        # Create an environment file for accessing the public API as the admin
        # user.
        playbooks = _build_playbook_list("public-openrc")
        self.run_kayobe_playbooks(parsed_args, playbooks, ignore_limit=True)


class OvercloudServiceStop(KollaAnsibleMixin, KayobeAnsibleMixin, VaultMixin,
                           Command):
    """Stop the overcloud services.

    * Configure kolla-ansible.
    * Configure overcloud services in kolla-ansible.
    * Perform a kolla-ansible stop of the overcloud services.
    * Stop kayobe extra services.

    This can be used in conjunction with the --tags and --kolla-tags arguments
    to stop specific services.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceStop, self).get_parser(prog_name)
        group = parser.add_argument_group("Services")
        group.add_argument("--yes-i-really-really-mean-it",
                           action='store_true',
                           help="confirm that you understand that this will "
                                "stop running services.")
        return parser

    def take_action(self, parsed_args):
        if not parsed_args.yes_i_really_really_mean_it:
            self.app.LOG.error("This will stop running services. Specify "
                               "--yes-i-really-really-mean-it to confirm that "
                               "you understand this.")
            sys.exit(1)

        self.app.LOG.debug("Stopping overcloud services")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args)

        # Perform the kolla-ansible stop.
        extra_args = ["--yes-i-really-really-mean-it"]
        self.run_kolla_ansible_overcloud(parsed_args, "stop",
                                         extra_args=extra_args)

        # Stop kayobe extra services.
        playbooks = _build_playbook_list("overcloud-extras")
        extra_vars = {"kayobe_action": "stop"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars, limit="overcloud")


class OvercloudServiceUpgrade(KollaAnsibleMixin, KayobeAnsibleMixin,
                              VaultMixin, Command):
    """Upgrade the overcloud services.

    * Configure kolla-ansible.
    * Configure overcloud services in kolla-ansible.
    * Perform kolla-ansible prechecks to verify the system state for
      deployment.
    * Perform a kolla-ansible upgrade of the overcloud services.
    * Configure and upgrade kayobe extra services.
    * Regenerate openrc files for the admin user.

    This can be used in conjunction with the --tags and --kolla-tags arguments
    to upgrade specific services.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceUpgrade, self).get_parser(prog_name)
        group = parser.add_argument_group("Service Upgrade")
        group.add_argument("--skip-prechecks", action='store_true',
                           help="skip the kolla-ansible prechecks command")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Upgrading overcloud services")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args, install=True)

        # Run kolla-ansible prechecks before upgrade.
        if not parsed_args.skip_prechecks:
            self.run_kolla_ansible_overcloud(parsed_args, "prechecks")

        # Perform the kolla-ansible upgrade.
        self.run_kolla_ansible_overcloud(parsed_args, "upgrade")

        # Upgrade kayobe extra services.
        playbooks = _build_playbook_list("overcloud-extras")
        extra_vars = {"kayobe_action": "upgrade"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars, limit="overcloud")

        # Post-deployment configuration.
        self.run_kolla_ansible_overcloud(parsed_args, "post-deploy")
        # Create an environment file for accessing the public API as the admin
        # user.
        playbooks = _build_playbook_list("public-openrc")
        self.run_kayobe_playbooks(parsed_args, playbooks, ignore_limit=True)


class OvercloudServiceDestroy(KollaAnsibleMixin, KayobeAnsibleMixin,
                              VaultMixin, Command):
    """Destroy the overcloud services.

    Permanently destroy the overcloud containers, container images, and
    container volumes.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudServiceDestroy, self).get_parser(prog_name)
        group = parser.add_argument_group("Services")
        group.add_argument("--yes-i-really-really-mean-it",
                           action='store_true',
                           help="confirm that you understand that this will "
                                "permantently destroy all services and data.")
        return parser

    def take_action(self, parsed_args):
        if not parsed_args.yes_i_really_really_mean_it:
            self.app.LOG.error("This will permanently destroy all services "
                               "and data. Specify "
                               "--yes-i-really-really-mean-it to confirm that "
                               "you understand this.")
            sys.exit(1)

        self.app.LOG.debug("Destroying overcloud services")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args)

        # Run kolla-ansible destroy.
        extra_args = ["--yes-i-really-really-mean-it"]
        self.run_kolla_ansible_overcloud(parsed_args, "destroy",
                                         extra_args=extra_args)

        # Destroy kayobe extra services.
        playbooks = _build_playbook_list("overcloud-extras")
        extra_vars = {"kayobe_action": "destroy"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars, limit="overcloud")


class OvercloudContainerImagePull(KayobeAnsibleMixin, KollaAnsibleMixin,
                                  VaultMixin, Command):
    """Pull the overcloud container images from a registry."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Pulling overcloud container images")

        # First prepare configuration.
        self.generate_kolla_ansible_config(parsed_args, service_config=False)

        # Pull updated kolla container images.
        self.run_kolla_ansible_overcloud(parsed_args, "pull")

        # Pull container images for kayobe extra services.
        playbooks = _build_playbook_list("overcloud-extras")
        extra_vars = {"kayobe_action": "pull"}
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars, limit="overcloud")


class OvercloudContainerImageBuild(KayobeAnsibleMixin, VaultMixin, Command):
    """Build the overcloud container images."""

    def get_parser(self, prog_name):
        parser = super(OvercloudContainerImageBuild, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Container Image Build")
        group.add_argument("--nocache", action="store_true",
                           help="whether to not use cache")
        group.add_argument("--push", action="store_true",
                           help="whether to push images to a registry after "
                                "building")
        group.add_argument("regex", nargs='*',
                           help="regular expression matching names of images "
                                "to build. Builds all images if unspecified")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Building overcloud container images")
        playbooks = _build_playbook_list(
            "container-image-builders-check", "kolla-build",
            "container-image-build")
        extra_vars = {
            "nocache": parsed_args.nocache,
            "push_images": parsed_args.push
        }
        if parsed_args.regex:
            regexes = " ".join(parsed_args.regex)
            extra_vars["container_image_regexes"] = regexes
        else:
            extra_vars["container_image_sets"] = (
                "{{ overcloud_container_image_sets }}")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class OvercloudDeploymentImageBuild(KayobeAnsibleMixin, VaultMixin, Command):
    """Build the overcloud deployment kernel and ramdisk images."""

    def get_parser(self, prog_name):
        parser = super(OvercloudDeploymentImageBuild, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Deployment Image Build")
        group.add_argument("--force-rebuild", action="store_true",
                           help="whether to force rebuilding the images")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Building overcloud deployment images")
        playbooks = _build_playbook_list("overcloud-ipa-build")
        extra_vars = {}
        if parsed_args.force_rebuild:
            extra_vars["ipa_image_force_rebuild"] = True
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class OvercloudHostImageBuild(KayobeAnsibleMixin, VaultMixin, Command):
    """Build overcloud host disk images.

    Builds host disk images using Diskimage Builder (DIB) for use when
    provisioning the overcloud hosts.
    """

    def get_parser(self, prog_name):
        parser = super(OvercloudHostImageBuild, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Host Image Build")
        group.add_argument("--force-rebuild", action="store_true",
                           help="whether to force rebuilding the images")
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Building overcloud host disk images")
        playbooks = _build_playbook_list("overcloud-host-image-build")
        extra_vars = {}
        if parsed_args.force_rebuild:
            extra_vars["overcloud_host_image_force_rebuild"] = True
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class OvercloudPostConfigure(KayobeAnsibleMixin, VaultMixin, Command):
    """Perform post-deployment configuration.

    * Register Ironic Python Agent (IPA) deployment images using Diskimage
      Builder (DIB), if building deployment images locally.
    * Register ironic inspector introspection rules with the overcloud
      inspector service.
    * Register a provisioning network with neutron.
    * Configure serial consoles for the ironic nodes
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Performing post-deployment configuration")
        playbooks = _build_playbook_list(
            "overcloud-ipa-images", "overcloud-introspection-rules",
            "overcloud-introspection-rules-dell-lldp-workaround",
            "provision-net", "baremetal-compute-serial-console-post-config")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class OvercloudSwiftRingsGenerate(KayobeAnsibleMixin, VaultMixin, Command):
    """Generate Swift rings."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Generating Swift rings")
        playbooks = _build_playbook_list("swift-rings")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class NetworkConnectivityCheck(KayobeAnsibleMixin, VaultMixin, Command):
    """Check network connectivity between hosts in the control plane.

    Checks for access to an external IP address, an external hostname, any
    configured gateways, and between hosts on the same subnets. The MTU of
    each network is validated by sending ping packets of maximum size.
    """

    def take_action(self, parsed_args):
        self.app.LOG.debug("Performing network connectivity check")
        playbooks = _build_playbook_list("network-connectivity")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class BaremetalComputeRegister(KayobeAnsibleMixin, VaultMixin, Command):
    """Register baremetal compute nodes in Ironic."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Register baremetal compute nodes in Ironic.")
        playbooks = _build_playbook_list("baremetal-compute-register")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class BaremetalComputeInspect(KayobeAnsibleMixin, VaultMixin, Command):
    """Perform hardware inspection on baremetal compute nodes."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Performing hardware inspection on baremetal "
                           "compute nodes")
        playbooks = _build_playbook_list("baremetal-compute-inspect")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class BaremetalComputeIntrospectionDataSave(KayobeAnsibleMixin, VaultMixin,
                                            Command):
    """Save hardware introspection data for the baremetal compute nodes.

    Save hardware introspection data from the overcloud's ironic inspector
    service to the Ansible control host.
    """

    def get_parser(self, prog_name):
        parser = super(BaremetalComputeIntrospectionDataSave, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Introspection data")
        # Defaults for these are applied in the playbook.
        group.add_argument("--output-dir", type=str,
                           help="Path to directory in which to save "
                                "introspection data. Default: "
                                "$PWD/baremetal-compute-introspection-data")
        group.add_argument("--output-format", type=str,
                           help="Format in which to save output data. One of "
                                "JSON or YAML. Default: JSON",
                           choices=["JSON", "YAML"])
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Saving introspection data")
        extra_vars = {}
        if parsed_args.output_dir:
            extra_vars['output_dir'] = parsed_args.output_dir
        if parsed_args.output_format:
            extra_vars['output_format'] = parsed_args.output_format
        playbooks = _build_playbook_list(
            "baremetal-compute-introspection-data-save")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class BaremetalComputeManage(KayobeAnsibleMixin, VaultMixin, Command):
    """Put baremetal compute nodes into the manageable provision state."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Making baremetal compute nodes manageable")
        playbooks = _build_playbook_list("baremetal-compute-manage")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class BaremetalComputeProvide(KayobeAnsibleMixin, VaultMixin, Command):
    """Put baremetal compute nodes into the available provision state."""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Making baremetal compute nodes available")
        playbooks = _build_playbook_list("baremetal-compute-provide")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class BaremetalComputeRename(KayobeAnsibleMixin, VaultMixin, Command):
    """Rename baremetal compute nodes to match inventory hostname"""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Renaming baremetal compute nodes")
        playbooks = _build_playbook_list("baremetal-compute-rename")
        self.run_kayobe_playbooks(parsed_args, playbooks)


class BaremetalComputeSerialConsoleBase(KayobeAnsibleMixin, VaultMixin,
                                        Command):

    """Base class for the baremetal serial console commands"""

    @staticmethod
    def process_limit(parsed_args, extra_vars):
        if parsed_args.baremetal_compute_limit:
            extra_vars["console_compute_node_limit"] = (
                parsed_args.baremetal_compute_limit
            )

    def get_parser(self, prog_name):
        parser = super(BaremetalComputeSerialConsoleBase, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Baremetal Serial Consoles")
        group.add_argument("--baremetal-compute-limit",
                           help="Limit the change to the hosts specified in "
                                "this limit"
                           )
        return parser


class BaremetalComputeSerialConsoleEnable(BaremetalComputeSerialConsoleBase):
    """Enable Serial Console for Baremetal Compute Nodes"""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Enabling serial console for ironic nodes")
        extra_vars = {}
        BaremetalComputeSerialConsoleBase.process_limit(parsed_args,
                                                        extra_vars)
        extra_vars["cmd"] = "enable"
        playbooks = _build_playbook_list("baremetal-compute-serial-console")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class BaremetalComputeSerialConsoleDisable(BaremetalComputeSerialConsoleBase):
    """Disable Serial Console for Baremetal Compute Nodes"""

    def take_action(self, parsed_args):
        self.app.LOG.debug("Disable serial console for ironic nodes")
        extra_vars = {}
        BaremetalComputeSerialConsoleBase.process_limit(parsed_args,
                                                        extra_vars)
        extra_vars["cmd"] = "disable"
        playbooks = _build_playbook_list("baremetal-compute-serial-console")
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class BaremetalComputeUpdateDeploymentImage(KayobeAnsibleMixin, VaultMixin,
                                            Command):
    """Update the Ironic nodes to use the new kernel and ramdisk images."""

    def get_parser(self, prog_name):
        parser = super(BaremetalComputeUpdateDeploymentImage, self).get_parser(
            prog_name)
        group = parser.add_argument_group("Baremetal Compute Update")
        group.add_argument("--baremetal-compute-limit",
                           help="Limit the upgrade to the hosts specified in "
                                "this limit"
                           )
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug(
            "Upgrading the ironic nodes to use the latest deployment images")
        playbooks = _build_playbook_list("overcloud-ipa-images")
        extra_vars = {}
        extra_vars["ipa_images_update_ironic_nodes"] = True
        if parsed_args.baremetal_compute_limit:
            extra_vars["ipa_images_compute_node_limit"] = (
                parsed_args.baremetal_compute_limit
            )
        self.run_kayobe_playbooks(parsed_args, playbooks,
                                  extra_vars=extra_vars)


class EnvironmentCreate(KayobeAnsibleMixin, VaultMixin, Command):
    """Create a new Kayobe environment."""

    def get_parser(self, prog_name):
        parser = super(EnvironmentCreate, self).get_parser(prog_name)
        group = parser.add_argument_group("Kayobe Environments")
        environment.add_args(group)
        return parser

    def take_action(self, parsed_args):
        self.app.LOG.debug("Creating new Kayobe environment")
        if not parsed_args.environment:
            self.app.LOG.error("An environment must be specified")
            sys.exit(1)

        source_config_path = parsed_args.source_config_path
        if source_config_path:
            result = utils.is_readable_dir(source_config_path)
            if not result["result"]:
                self.app.LOG.error("Kayobe configuration %s is invalid: %s",
                                   source_config_path, result["message"])
                sys.exit(1)

        try:
            environment.create_kayobe_environment(parsed_args)
        except Exception as e:
            self.app.LOG.error("Failed to create environment %s: %s",
                               parsed_args.environment, repr(e))
            sys.exit(1)



#!/usr/bin/env python

# Copyright 2018 StackHPC Ltd.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ansible.errors import AnsibleModuleError
from oslotest import base

from kayobe.plugins.action import merge_yaml


class MergeYamlConfigTest(base.BaseTestCase):

    def test_merge_no_update(self):
        initial_conf = {
            'foo': 'bar',
            'egg': 'spam'
        }
        actual = merge_yaml.Utils.update_nested_conf(initial_conf, {})
        expected = {
            'foo': 'bar',
            'egg': 'spam'
        }
        self.assertDictEqual(actual, expected)

    def test_merge_flat_update_key(self):
        initial_conf = {
            'foo': 'bar',
            'egg': 'spam'
        }
        actual = merge_yaml.Utils.update_nested_conf(
            initial_conf, {'egg': 'ham'})
        expected = {
            'foo': 'bar',
            'egg': 'ham'
        }
        self.assertDictEqual(actual, expected)

    def test_merge_flat_new_key(self):
        initial_conf = {
            'foo': 'bar',
            'egg': 'spam'
        }
        actual = merge_yaml.Utils.update_nested_conf(
            initial_conf, {'spam': 'ham'})
        expected = {
            'foo': 'bar',
            'egg': 'spam',
            'spam': 'ham'
        }
        self.assertDictEqual(actual, expected)

    def test_merge_nested_update_key(self):
        initial_conf = {
            'foo': {
                'a': 'b',
            },
            'bar': {
                'a': False,
                'b': 'INFO'
            }
        }
        actual = merge_yaml.Utils.update_nested_conf(
            initial_conf, {'bar': {'a': True}})
        expected = {
            'foo': {
                'a': 'b',
            },
            'bar': {
                'a': True,
                'b': 'INFO'
            }
        }
        self.assertDictEqual(actual, expected)

    def test_merge_nested_new_key(self):
        initial_conf = {
            'foo': {
                'a': 'b',
                'c': 30
            }
        }
        actual = merge_yaml.Utils.update_nested_conf(
            initial_conf, {'egg': {'spam': 10}})
        expected = {
            'foo': {
                'a': 'b',
                'c': 30,
            },
            'egg': {
                'spam': 10,
            }
        }
        self.assertDictEqual(actual, expected)

    def test_merge_nested_new_nested_key(self):
        initial_conf = {
            'foo': {
                'a': 'b',
                'c': 30
            }
        }
        actual = merge_yaml.Utils.update_nested_conf(
            initial_conf, {'foo': {'spam': 10}})
        expected = {
            'foo': {
                'a': 'b',
                'c': 30,
                'spam': 10,
            }
        }
        self.assertDictEqual(actual, expected)

    def test_merge_nested_extend_lists(self):
        initial_conf = {
            'level0': {
                'level1': {
                    "mylist": ["one", "two"]
                },
            }
        }

        extension = {
            'level0': {
                'level1': {
                    "mylist": ["three"]
                },
            }
        }

        actual = merge_yaml.Utils.update_nested_conf(
            initial_conf, extension, extend_lists=True)
        expected = {
            'level0': {
                'level1': {
                    "mylist": ["one", "two", "three"]
                },
            }
        }
        self.assertDictEqual(actual, expected)

    def test_merge_nested_extend_lists_mismatch_types(self):
        initial_conf = {
            'level0': {
                'level1': {
                    "mylist": ["one", "two"]
                },
            }
        }

        extension = {
            'level0': {
                'level1': {
                    "mylist": "three"
                },
            }
        }
        with self.assertRaisesRegex(AnsibleModuleError, "Failure merging key"):
            merge_yaml.Utils.update_nested_conf(
                initial_conf, extension, extend_lists=True)



DOCUMENTATION = '''
---
module: blockdevice_info
short_description: Returns information about block devices
version_added: "N/A"
description:
    - "Returns information about block devices"
author:
    - Will Szumski
'''

EXAMPLES = '''
- name: Retrieve information about block devices
  blockdevice_info:
  become: true
  register: result
'''

RETURN = '''
umounted:
    description: A list of all umounted block devices.
    type: list
    returned: always
'''

import json

from ansible.module_utils.basic import AnsibleModule

def _has_mounts(device):
    try:
        if device["mountpoint"]:
            return True
    # If unmounted, the JSON output contains "mountpoints": [null] so we handle
    # the KeyError here.
    except KeyError:
        if device["mountpoints"][0]:
            return True
    for child in device.get("children", []):
        if _has_mounts(child):
            return True
    return False

def unmounted(module, lsblk):
    result = []
    for device in lsblk.get("blockdevices", []):
        if not _has_mounts(device) and device["type"] == 'disk':
            result.append(device["name"])
    return result

def run_module():
    # The module takes no argumnets.
    module_args = dict()

    result = dict(
        changed=False,
        unmounted=[]
    )

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True
    )

    _rc, stdout, _stderr = module.run_command("lsblk -J")
    lsblk = json.loads(stdout)

    result['unmounted'] = unmounted(module, lsblk)

    module.exit_json(**result)


def main():
    run_module()


if __name__ == '__main__':
    main()



# Copyright (c) 2021 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import unittest

import jinja2

from ansible import errors
from kayobe.plugins.filter import networks


class BaseNetworksTest(unittest.TestCase):

    maxDiff = 2000

    variables = {
        # Inventory hostname, used to index IP list.
        "inventory_hostname": "test-host",
        # net1: Ethernet on eth0 with IP 1.2.3.4/24.
        "net1_interface": "eth0",
        # net2: VLAN on eth0.2 with VLAN 2 on interface eth0.
        "net2_interface": "eth0.2",
        "net2_vlan": 2,
        # net3: bridge on br0 with ports eth0 and eth1.
        "net3_interface": "br0",
        "net3_bridge_ports": ['eth0', 'eth1'],
        # net4: VLAN on br0.4 with VLAN 4 on bridge br0.
        "net4_interface": "br0.4",
        "net4_vlan": 4,
        # net5: VLAN on br0.5 with VLAN 5 on bridge br0.
        "net5_interface": "br0.5",
        "net5_vlan": 5,
        # Veth pair patch link prefix and suffix.
        "network_patch_prefix": "p-",
        "network_patch_suffix_ovs": "-ovs",
        "network_patch_suffix_phy": "-phy",
    }

    def setUp(self):
        # Bandit complains about Jinja2 autoescaping without nosec.
        self.env = jinja2.Environment()  # nosec
        self.context = self._make_context(self.variables)

    def _make_context(self, parent):
        return self.env.context_class(
            self.env, parent=parent, name='dummy', blocks={})

    def _update_context(self, variables):
        updated_vars = copy.deepcopy(self.variables)
        updated_vars.update(variables)
        self.context = self._make_context(updated_vars)


class TestNetworks(BaseNetworksTest):

    def test_get_ovs_veths_empty(self):
        veths = networks.get_ovs_veths(self.context, [], None)
        self.assertEqual([], veths)

    def test_get_ovs_veths_eth(self):
        # Ethernet does not need a veth pair.
        self._update_context({"external_net_names": ["net1"]})
        veths = networks.get_ovs_veths(self.context, ["net1"], None)
        self.assertEqual([], veths)

    def test_get_ovs_veths_eth_vlan(self):
        # VLAN on Ethernet does not need a veth pair.
        self._update_context({"external_net_names": ["net2"]})
        veths = networks.get_ovs_veths(self.context, ["net2"], None)
        self.assertEqual([], veths)

    def test_get_ovs_veths_bridge(self):
        # Bridge needs a veth pair.
        self._update_context({"external_net_names": ["net3"]})
        veths = networks.get_ovs_veths(self.context, ["net3"], None)
        expected = [
            {
                "name": "p-br0-phy",
                "peer": "p-br0-ovs",
                "bridge": "br0",
                "mtu": None,
            }
        ]
        self.assertEqual(expected, veths)

    def test_get_ovs_veths_bridge_provision_wl(self):
        # Bridge needs a veth pair.
        self._update_context({"provision_wl_net_name": "net3"})
        veths = networks.get_ovs_veths(self.context, ["net3"], None)
        expected = [
            {
                "name": "p-br0-phy",
                "peer": "p-br0-ovs",
                "bridge": "br0",
                "mtu": None,
            }
        ]
        self.assertEqual(expected, veths)

    def test_get_ovs_veths_bridge_cleaning(self):
        # Bridge needs a veth pair.
        self._update_context({"cleaning_net_name": "net3"})
        veths = networks.get_ovs_veths(self.context, ["net3"], None)
        expected = [
            {
                "name": "p-br0-phy",
                "peer": "p-br0-ovs",
                "bridge": "br0",
                "mtu": None,
            }
        ]
        self.assertEqual(expected, veths)

    def test_get_ovs_veths_bridge_mtu(self):
        # Use the MTU of bridge.
        self._update_context({"external_net_names": ["net3"],
                              "net3_mtu": 1400})
        veths = networks.get_ovs_veths(self.context, ["net3"], None)
        expected = [
            {
                "name": "p-br0-phy",
                "peer": "p-br0-ovs",
                "bridge": "br0",
                "mtu": 1400,
            }
        ]
        self.assertEqual(expected, veths)

    def test_get_ovs_veths_bridge_vlan(self):
        # VLAN on a bridge needs a veth pair.
        self._update_context({"external_net_names": ["net4"]})
        veths = networks.get_ovs_veths(self.context, ["net3", "net4"], None)
        expected = [
            {
                "name": "p-br0-phy",
                "peer": "p-br0-ovs",
                "bridge": "br0",
                "mtu": None,
            }
        ]
        self.assertEqual(expected, veths)

    def test_get_ovs_veths_bridge_vlan_mtu(self):
        # Use the MTU of VLAN on a bridge.
        self._update_context({"external_net_names": ["net4"],
                              "net4_mtu": 1400})
        veths = networks.get_ovs_veths(self.context, ["net3", "net4"], None)
        expected = [
            {
                "name": "p-br0-phy",
                "peer": "p-br0-ovs",
                "bridge": "br0",
                "mtu": 1400,
            }
        ]
        self.assertEqual(expected, veths)

    def test_get_ovs_veths_bridge_vlan_multiple(self):
        # Multiple VLANs on a bridge need a single veth pair.
        self._update_context({"external_net_names": ["net4", "net5"]})
        veths = networks.get_ovs_veths(self.context, ["net3", "net4", "net5"],
                                       None)
        expected = [
            {
                "name": "p-br0-phy",
                "peer": "p-br0-ovs",
                "bridge": "br0",
                "mtu": None,
            }
        ]
        self.assertEqual(expected, veths)

    def test_get_ovs_veths_bridge_vlan_multiple_mtu(self):
        # Use the highest MTU of multiple VLANs on a bridge.
        self._update_context({"external_net_names": ["net4", "net5"],
                              "net4_mtu": 1400,
                              "net5_mtu": 1500})
        veths = networks.get_ovs_veths(self.context, ["net3", "net4", "net5"],
                                       None)
        expected = [
            {
                "name": "p-br0-phy",
                "peer": "p-br0-ovs",
                "bridge": "br0",
                "mtu": 1500,
            }
        ]
        self.assertEqual(expected, veths)

    def test_ensure_bridge_ports_is_list(self):
        self._update_context({"net3_bridge_ports": "ens3"})
        self.assertRaises(errors.AnsibleFilterError, networks.net_bridge_ports,
                          self.context, "net3")

    def test_physical_interface_bond(self):
        self._update_context({"net6_interface": "bond0",
                              "net6_bond_slaves": ["eth3", "eth4"]})
        interface = networks.net_physical_interface(self.context, "net6")
        expected = ['eth3', 'eth4']
        self.assertEqual(expected, interface)

    def test_physical_interface_bridge(self):
        interface = networks.net_physical_interface(self.context, "net3")
        expected = ['eth0', 'eth1']
        self.assertEqual(expected, interface)

    def test_physical_interface_direct(self):
        interface = networks.net_physical_interface(self.context, "net1")
        expected = ['eth0']
        self.assertEqual(expected, interface)
